# dbconnect.py

# dbconnect.py

import psycopg2
import os
from psycopg2 import extras, errors, pool  # IntegrityError
import pandas as pd
from pandas.io.sql import DatabaseError

DB_PARAMETERS = {
    "host": os.environ.get("DB_SERVER", ""),
    "port": int(os.environ.get("DB_PORT", "5432")),
    "database": os.environ.get("POSTGRES_DB", ""),
    "user": os.environ.get("POSTGRES_USER", ""),
    "password": os.environ.get("POSTGRES_PASSWORD", ""),
    "cursor_factory": extras.RealDictCursor,
}

assert len(DB_PARAMETERS["password"]) > 2, "Invalid DB connection password"

threaded_postgreSQL_pool = pool.ThreadedConnectionPool(5, 500, **DB_PARAMETERS)
assert threaded_postgreSQL_pool, "Could not create DB connection"

print("DB Connected")


def makeQuery(
    s1,
    output="oneValue",
    lowerCaseColumns=False,
    fillna=True,
    engine=None,
    noprint=False,
):
    """
    output choices:
    oneValue : ex: select count(*) from table1 (output will be only one value)
    df: ex: select * from users (output will be a table)
    list: json array, like df.to_dict(orient='records')
    column: first column in output as a list. ex: select username from users
    oneJson: First row, as dict
    """
    if not isinstance(s1, str):
        print("query needs to be a string")
        return False
    if ";" in s1:
        print("; not allowed")
        return False

    if not noprint:
        # keeping auth check and some other queries out
        skipPrint = ["where token=", ".STArea()", "STGeomFromText"]
        if not any([(x in s1) for x in skipPrint]):
            print(f"Query: {' '.join(s1.split())}")
        else:
            print(f"Query: {' '.join(s1.split())[:20]}")

    ps_connection = threaded_postgreSQL_pool.getconn()

    result = None  # default return value

    if output in ("oneValue", "oneJson"):
        ps_cursor = ps_connection.cursor()
        ps_cursor.execute(s1)
        row = ps_cursor.fetchone()
        ps_cursor.close()
        if not row:
            result = None
        else:
            if output == "oneValue":
                result = list(row.values())[0]
            else:
                result = dict(row)

    elif output in ("df", "list", "column"):
        try:
            ps_cursor = ps_connection.cursor()
            ps_cursor.execute(s1)
            res = ps_cursor.fetchall()
            ps_cursor.close()
            
            # res is now a dict array only, right?
            if output=="list":
                try:
                    threaded_postgreSQL_pool.putconn(
                        ps_connection
                    )  # return threaded connnection back to pool
                except DatabaseError as e:
                    print("DatabaseError when returning threaded connection back to pool")
                    print(e)
                finally:
                    return res

            if fillna:
                df = pd.DataFrame(res).fillna("")
            else:
                df = pd.DataFrame(res)
        except DatabaseError as e:
            print("DatabaseError!")
            print(e)
            raise

        # make all colunm headers lowercase
        if lowerCaseColumns:
            df.columns = [
                x.lower() for x in df.columns
            ]  # from https://stackoverflow.com/questions/19726029/how-can-i-make-pandas-dataframe-column-headers-all-lowercase

        if output == "df":
            result = df
            if not len(df):
                result = []

        elif not len(df):
            result = []
        elif output == "column":
            result = df.iloc[:, 0].tolist()  # .iloc[:,0] -> first column
        # elif output == "list":
        #     result = df.to_dict(orient="records")
        else:
            # default - df
            result = df
    else:
        print("invalid output type")

    try:
        threaded_postgreSQL_pool.putconn(
            ps_connection
        )  # return threaded connnection back to pool
    except DatabaseError as e:
        print("DatabaseError when returning threaded connection back to pool")
        print(e)
    finally:
        return result


def execSQL(s1, noprint=False):
    # if not noprint: print(' '.join(s1.split())[:200])
    if not noprint:
        print(" ".join(s1.split()))
    ps_connection = threaded_postgreSQL_pool.getconn()
    ps_cursor = ps_connection.cursor()
    try:
        ps_cursor.execute(s1)
        ps_connection.commit()

    except errors.UniqueViolation as e:
        print(f"UniqueViolation! {e}")
        return False

    affected = ps_cursor.rowcount
    ps_cursor.close()
    threaded_postgreSQL_pool.putconn(ps_connection)
    return affected


def addRow(params, tablename):
    df = pd.DataFrame([params])
    return addTable(df, tablename)  # heck


def addTable(df, table, noprint=False):
    """
    From https://naysan.ca/2020/05/09/pandas-to-postgresql-using-psycopg2-bulk-insert-performance-benchmark/
    Using psycopg2.extras.execute_values() to insert the dataframe
    https://www.psycopg.org/docs/extras.html#fast-exec
    """
    # Create a list of tupples from the dataframe values
    tuples = [tuple(x) for x in df.to_numpy()]
    # Comma-separated dataframe columns
    cols = ",".join(list(df.columns))
    # SQL query to execute
    query = "INSERT INTO %s(%s) VALUES %%s" % (table, cols)
    ps_connection = threaded_postgreSQL_pool.getconn()
    cursor = ps_connection.cursor()
    if not noprint: print(f"Adding {len(df)} rows to {table}")
    try:
        extras.execute_values(cursor, query, tuples)
        ps_connection.commit()
    except (Exception, psycopg2.DatabaseError) as error:
        print("Error: %s" % error)
        ps_connection.rollback()
        cursor.close()
        threaded_postgreSQL_pool.putconn(
            ps_connection
        )  # return threaded connnection back to pool
        return False
    cursor.close()
    threaded_postgreSQL_pool.putconn(
        ps_connection
    )  # return threaded connnection back to pool
    return True

